#include <itkVectorInterpolateImageFunction.h>
#include <itkVelocityFieldBCHCompositionFilter.h>
#include <itkImage.h>
#include <itkImageFileReader.h>
#include <itkImageFileWriter.h>
#include "itkDivideByConstantImageFilter.h"
#include <itkImageRegionIterator.h>
#include <itkAddImageFilter.h>
#include <itkVelocityFieldLieBracketFilter.h>
#include <itkWarpVectorImageFilter.h>
#include"math.h"
#include "itkMultiplyByConstantImageFilter.h"
#include "tclap/CmdLine.h"


/** 
  * This program computes the parallel transport of a stationary velocity field (SVF) X along the direction of a SVF Y. 
  * The parallel transport is implemented by the Schild's Ladder, which transports a vector X by iterative computation of geodesic parallelograms along the direction parametrized by Y   
  * It can be shown that  
  * \Pi^{tY}(X)\simeq X+t[Y,X]+t^2[Y,[Y,X]]
  * For details see M. Lorenzi, N Ayache and X. Pennec, IPMI 2011. 
  **/  


typedef itk::Vector <float,3> VectorType;
typedef itk::Image <VectorType,3> VelocityFieldType;
typedef VelocityFieldType::Pointer VelocityFieldPointer;  

typedef itk::Image<float,3> ScalarImageType;
typedef ScalarImageType::Pointer ScalarImagePointer;



typedef itk::Vector<float,3> PixelType;
typedef itk::Image<PixelType,3> ImageType;
typedef ImageType::Pointer InputFieldPointer;

struct Param{
	    std::string  DirectionImagePath;
	    std::string  VectorToTransportImagePath;
	    std::string  OutputImagePath;
	    float ScalingFactor;
	    std::string  MaskImagePath;
};


/**
  * Parses the command line arguments and deduces the corresponding Param structure.
  * @param  argc   number of arguments
  * @param  argv   array containing the arguments
  * @param  param  structure of parameters
  */

void parseParameters(int argc, char** argv, struct Param & param)
{

    // Program description
    std::string description = "\b\b\bDESCRIPTION\n";
    description += "Schild's Ladder for Stationary Velocity Fields";
    description += "\nAuthor : Marco Lorenzi";
    try {
	 // Define the command line parser
         TCLAP::CmdLine cmd( description, ' ', "1.0", true);
         TCLAP::ValueArg<std::string>  arg_DirectionImage( "d", "direction-svf", "Path to the stationary velocity field defining the direction of transport", true, "", "string", cmd );
	 TCLAP::ValueArg<std::string>  arg_OutputImage( "t", "output-svf", "Path of the transported stationary velocity field (default output.mha).", false, "output.mha", "string", cmd );
	 TCLAP::ValueArg<std::string>  arg_VectorToTransportImage( "v", "transported-svf", "Path to the stationary velocity field to be transported", true, "", "string", cmd );
	 TCLAP::ValueArg<double>  arg_ScalingFactor( "s", "scaling-factor", "Scaling factor for the direction of transport (default 1)", false, 1.0, "double", cmd );
	 TCLAP::ValueArg<std::string>  arg_MaskImage( "m", "mask", "Path to the mask", true, "", "string", cmd );
	 
	 // Parse the command line
	 cmd.parse( argc, argv );
	 // Set the parameters
         param.DirectionImagePath           = arg_DirectionImage.getValue();
         param.VectorToTransportImagePath   = arg_VectorToTransportImage.getValue();
         param.OutputImagePath              = arg_OutputImage.getValue();
         param.ScalingFactor                = arg_ScalingFactor.getValue();
	 param.MaskImagePath                = arg_MaskImage.getValue();
        }
    catch (TCLAP::ArgException &e)
        {
         std::cerr << "Error: " << e.error() << " for argument " << e.argId() << std::endl;
         throw std::runtime_error("Unable to parse the command line arguments.");
        }
}

/**
  * Prints parameters.
  * @param  DirectionImagePath              path to the direction vector
  * @param  VectorToTransportImagePath      path to the transported vector
  * @param  OutputImagePath                 path to the output file containing the transported vector
  * @param  Scaling Factor                  ScalingFactor
  * @param  MaskImagePath                   path to the mask
  **/
void PrintParameters( std::string DirectionImagePath,
	              std::string VectorToTransportImagePath,
	              std::string OutputImagePath,
		      double      ScalingFactor, 
		      std::string MaskImagePath)
{

       // Print I/O parameters
       std::cout << std::endl;
       std::cout << "I/O PARAMETERS"                  << std::endl;
       std::cout << "  Direction image path           : " << DirectionImagePath      << std::endl;
       std::cout << "  VectorToTransport image path   : " << VectorToTransportImagePath     << std::endl;
       std::cout << "  Output image path              : " << OutputImagePath     << std::endl;
       std::cout << "  ScalingFactor                  : " << ScalingFactor << std::endl;
       std::cout << "  Mask image path                : " << MaskImagePath     << std::endl;
       std::cout << std::endl;
}



















float GetScalingFactor(VelocityFieldPointer InputSVF,ScalarImagePointer Mask )
  {
  double minpixelspacing = InputSVF->GetSpacing()[0];
    for (unsigned int i = 0; i<3; ++i)
      {
      if ( InputSVF->GetSpacing()[i] < minpixelspacing/2 )
        {
        minpixelspacing = InputSVF->GetSpacing()[i];
        }
      }

   typedef itk::ImageRegionIterator<VelocityFieldType> IteratorType;  

    IteratorType InputIt (InputSVF, InputSVF->GetRequestedRegion());
 
 typedef itk::Index<3> indexType;

   float norm2,maxnorm2=0;
    for( InputIt.GoToBegin(); !InputIt.IsAtEnd(); ++InputIt )
      {
      indexType index=InputIt.GetIndex();
       if ( (Mask->GetPixel(index))>0)
         { 
	 norm2 = InputIt.Get().GetSquaredNorm();
         if (maxnorm2<norm2) maxnorm2=norm2;
	 }
      }

    /*maxnorm2 /= vnl_math_sqr(minpixelspacing);
    std::cout<<"maxnorm(after):" <<maxnorm2<<std::endl;*/
   
    float numiterfloat = 2*pow(maxnorm2,0.5)/minpixelspacing;
   if (numiterfloat<1) numiterfloat=1; 
     
  return((abs(numiterfloat)));

  }


VelocityFieldPointer Resample (VelocityFieldPointer InputSVF, VelocityFieldPointer Deformation){

typedef itk::WarpVectorImageFilter<VelocityFieldType,VelocityFieldType,VelocityFieldType>  WarperType;

WarperType::Pointer warper=WarperType::New();
warper->SetInput(InputSVF);
warper->SetDeformationField(Deformation);
warper->SetOutputDirection(Deformation->GetDirection());
warper->SetOutputOrigin(Deformation->GetOrigin());
warper->SetOutputSpacing(Deformation->GetSpacing());
warper->Update();
return(warper->GetOutput());
}



VelocityFieldPointer Exponential(VelocityFieldPointer Vector, ScalarImagePointer Mask){

VelocityFieldPointer Out1=Vector;
double minpixelspacing = Vector->GetSpacing()[0];
    for (unsigned int i = 0; i<2; ++i)
      {
      if ( Out1->GetSpacing()[i] < minpixelspacing/2 )
        {
        minpixelspacing = Out1->GetSpacing()[i];
        }
      }

 typedef itk::ImageRegionIterator<VelocityFieldType> IteratorType;  
  
    IteratorType InputIt (Vector, Vector->GetRequestedRegion());
 typedef itk::Index<3> indexType;
  
   float norm2,maxnorm2=0;
    for( InputIt.GoToBegin(); !InputIt.IsAtEnd(); ++InputIt )
      { 
       indexType index=InputIt.GetIndex();
       if ( (Mask->GetPixel(index))>0)
         { 
	 norm2 = InputIt.Get().GetSquaredNorm();
         if (maxnorm2<norm2) maxnorm2=norm2;
	 }
      }
    maxnorm2 /= vnl_math_sqr(minpixelspacing);
    
    float numiterfloat = 2.0 +
       0.5 * vcl_log(maxnorm2)/vnl_math::ln2;
       
  typedef itk::DivideByConstantImageFilter<VelocityFieldType,float,VelocityFieldType> DividerType;
  DividerType::Pointer Divider=DividerType::New();
      Divider->SetInput(Out1);
   Divider->SetConstant( static_cast<unsigned int>(1<<static_cast<unsigned int>(abs(numiterfloat))) );
 
  Divider->Update();

  typedef itk::WarpVectorImageFilter<VelocityFieldType,VelocityFieldType,VelocityFieldType> WarperType;
    typedef itk::AddImageFilter<VelocityFieldType,VelocityFieldType,VelocityFieldType> AdderType;
  AdderType::Pointer Adder=AdderType::New();  
  WarperType::Pointer Warper=WarperType::New();
  Warper->SetOutputOrigin(Out1->GetOrigin());
  Warper->SetOutputSpacing(Out1->GetSpacing());
  Warper->SetOutputDirection(Out1->GetDirection());
  
  
   VelocityFieldType::Pointer Vect1=VelocityFieldType::New();
   VelocityFieldType::Pointer warpedIm=VelocityFieldType::New();
    Vect1=Divider->GetOutput();
 
   for( int i=0; i<abs(numiterfloat); i++ )
    {

    Warper->SetInput(Vect1);
    Warper->SetDeformationField(Vect1);

    Warper->GetOutput()->SetRequestedRegion(
       Vect1->GetRequestedRegion() );

    Warper->Update();

    warpedIm = Warper->GetOutput();
  
    Adder->SetInput1(Vect1);

    Adder->SetInput2(warpedIm);
    Adder->GetOutput()->SetRequestedRegion(
      Vect1->GetRequestedRegion() );

    Adder->Update();

    Vect1=Adder->GetOutput();
 
    
    
    }


return(Vect1);

}


VelocityFieldPointer BCH(VelocityFieldPointer Input1, VelocityFieldPointer Input2){

  typedef itk::VelocityFieldLieBracketFilter <VelocityFieldType, VelocityFieldType>   LieFilterType;
  LieFilterType::Pointer LieFilter1=LieFilterType::New();
  LieFilterType::Pointer LieFilter2=LieFilterType::New();

  typedef itk::MultiplyByConstantImageFilter <VelocityFieldType,float,VelocityFieldType> MultiplierType;
  MultiplierType::Pointer Multiplier1=MultiplierType::New();
  MultiplierType::Pointer Multiplier2=MultiplierType::New();

  typedef itk::AddImageFilter <VelocityFieldType,VelocityFieldType,VelocityFieldType> AdderType;
  AdderType::Pointer Adder1=AdderType::New();
  AdderType::Pointer Adder2=AdderType::New();

  LieFilter1->SetInput(0,Input1);
  LieFilter1->SetInput(1,Input2);
  LieFilter1->Update();
   
  
  Multiplier1->SetInput(LieFilter1->GetOutput());
  Multiplier1->SetConstant(0.5);
  Multiplier1->Update();

  Adder1->SetInput1(Multiplier1->GetOutput());
  Adder1->SetInput2(Input1);
  Adder1->Update();

  Adder2->SetInput1(Adder1->GetOutput());

  Adder2->SetInput2(Input2);
  Adder2->Update();

  
return(Adder2->GetOutput());
}



int main( int argc, char ** argv ){

  
  typedef itk::ImageFileReader <VelocityFieldType> ReaderType;
  typedef itk::ImageFileReader <ScalarImageType> ScalarReaderType;

  //Parsing initial parameters
  struct Param param;
  parseParameters( argc, argv, param);

  
  /* Reader 1: Vector to be transported
     Reader 2: Direction for the transport */
  ReaderType::Pointer reader1 = ReaderType::New();
  ReaderType::Pointer reader2 = ReaderType::New();
  ScalarReaderType::Pointer readerMask = ScalarReaderType::New();

  reader1->SetFileName(param.VectorToTransportImagePath);
  reader2->SetFileName(param.DirectionImagePath);
  readerMask->SetFileName(param.MaskImagePath);

  reader1->Update();
  reader2->Update();
  readerMask->Update();

  ScalarImagePointer Mask=readerMask->GetOutput();

  /*
   * Scaling the direction of transport
   */
  typedef itk::DivideByConstantImageFilter<VelocityFieldType,float,VelocityFieldType> DividerType;
  DividerType::Pointer Divider=DividerType::New();
  Divider->SetInput(reader2->GetOutput());
  float ScalingFactor=GetScalingFactor(reader2->GetOutput(), Mask);    
  Divider->SetConstant( -ScalingFactor );
  Divider->Update(); 

  VelocityFieldPointer Diagonal=Divider->GetOutput();//Diagonal is the scaled direction of transport

  Diagonal->DisconnectPipeline();


  /*
   * Scaling the vector to transport
   */
  unsigned int ScalingFactor1=GetScalingFactor(reader1->GetOutput(), Mask);
  Divider->SetInput(reader1->GetOutput());
  Divider->SetConstant( ScalingFactor1 );
  Divider->Update();

  VelocityFieldPointer Pivot=Divider->GetOutput();

  Pivot->DisconnectPipeline();//Pivot is the scaled vector to transport

 // VelocityFieldPointer Deformation= Exponential(Divider->GetOutput());
typedef itk::VelocityFieldLieBracketFilter<VelocityFieldType,VelocityFieldType> LieFilterType;
LieFilterType::Pointer LieFilter1=LieFilterType::New();
LieFilterType::Pointer LieFilter2=LieFilterType::New();

typedef itk::AddImageFilter <VelocityFieldType,VelocityFieldType,VelocityFieldType> AdderType;
  AdderType::Pointer Adder1=AdderType::New();
  AdderType::Pointer Adder2=AdderType::New();
  
typedef itk::MultiplyByConstantImageFilter<VelocityFieldType,double,VelocityFieldType> MultiplierType;
  MultiplierType::Pointer Multiplier1=MultiplierType::New();
  MultiplierType::Pointer Multiplier2=MultiplierType::New();



  for (unsigned int i =0;i<param.ScalingFactor*ScalingFactor;++i)
    { 
     LieFilter1->SetInput(0,Diagonal);
     LieFilter1->SetInput(1,Pivot);
     LieFilter1->Update();
   
     LieFilter2->SetInput(0,Diagonal);
     LieFilter2->SetInput(1,LieFilter1->GetOutput());
     LieFilter2->Update();

     Multiplier1->SetInput(LieFilter2->GetOutput());
     Multiplier1->SetConstant(0.5);
     Multiplier1->Update();

     Adder1->SetInput1(Multiplier1->GetOutput());
     Adder1->SetInput2(LieFilter1->GetOutput());
     Adder1->Update();

     Adder2->SetInput1(Adder1->GetOutput());

     Adder2->SetInput2(Pivot);
     Adder2->Update();
   
     Pivot=Adder2->GetOutput();
     //Diagonal=Resample(Diagonal,Exponential(Diagonal,Mask) );
     Pivot->DisconnectPipeline();
    //Diagonal->DisconnectPipeline(); 
   // Pivot->DisconnectPipeline();

    }	    

  MultiplierType::Pointer MultiplierFinal=MultiplierType::New();

    MultiplierFinal->SetInput(Pivot);
    MultiplierFinal->SetConstant(ScalingFactor1);
    MultiplierFinal->Update();    


 typedef itk::ImageFileWriter <VelocityFieldType> WriterType;
 
 WriterType::Pointer writer =WriterType::New();
 writer->SetInput(MultiplierFinal->GetOutput());
 writer->SetFileName(param.OutputImagePath);
 writer->Update();


return 0;
}
